require 'thread'
require 'abstractify'
require 'kiss/constants'
require 'kiss/kiss'

module Kiss
    class KissAbstract < Kiss
        include Abstractify::Abstract

        abstract :write_interface, :read_interface, :connect, :close

        protected
        def initialize(strip_df_start=true)
            @strip_df_start = strip_df_start
            @frame_buffer = []
            @lock = Mutex.new
        end

        private
        def self.strip_df_start(frame)
            while frame[0] == DATA_FRAME
                frame.shift
            end
            while frame[0] and frame[0].chr == ' '
                frame.shift
            end
            while frame[-1] and frame[-1].chr == ' '
                frame.pop
            end
            frame
        end

        private
        def self.escape_special_codes(raw_code_bytes)
            encoded_bytes = []
            raw_code_bytes.each do |raw_code_byte|
                if raw_code_byte == FESC
                    encoded_bytes += FESC_TFESC
                elsif raw_code_byte == FEND
                    encoded_bytes += FESC_TFEND
                else
                    encoded_bytes += [raw_code_byte]
                end
            end
            encoded_bytes
        end

        private
        def self.command_byte_combine(port, command_code)
            if port > 127 or port < 0
                raise 'port out of range'
            elsif command_code > 127 or command_code < 0
                raise 'command_Code out of range'
            end
            (port << 4) & command_code
        end

        private
        def self.extract_path(start, raw_frame)
            full_path = []

            (2...start).each do |i|
                path = identity_as_string(extract_callsign(raw_frame[i * 7..-1]))
                if path and path.length > 0
                    if raw_frame[i * 7 + 6] & 0x80 != 0
                        full_path << [path, '*'].join
                    else
                        full_path << path
                    end
                end
            end
            return full_path
        end

        private
        def self.parse_identity_string(identity_string)
            # If we are parsing a spent token then first lets get rid of the astresick suffix.
            if identity_string[-1] == '*'
                identity_string = identity_string[0..-1]
            end

            if identity_string.include? '-'
                call_sign, ssid = identity_string.split('-')
            else
                call_sign = identity_string
                ssid = 0
            end

            return {:callsign => call_sign, :ssid => ssid.to_i}
        end

        private
        def self.identity_as_string(identity)
            if identity[:ssid] and identity[:ssid] > 0
                return [identity[:callsign], identity[:ssid].to_s].join('-')
            else
                return identity[:callsign]
            end
        end

        private
        def self.extract_callsign(raw_frame)
            callsign_as_array = raw_frame[0...6].map { |x| (x >> 1).chr }
            callsign = callsign_as_array.join.strip
            ssid = (raw_frame[6] >> 1) & 0x0f
            ssid = (ssid == nil or ssid == 0 ? nil : ssid)
            return {:callsign => callsign, :ssid => ssid}
        end

        private
        def self.encode_callsign(callsign)
            call_sign = callsign[:callsign]

            enc_ssid = (callsign[:ssid] << 1) | 0x60

            if call_sign.include? '*'
                call_sign.gsub!(/\*/, '')
                enc_ssid |= 0x80
            end

            while call_sign.length < 6
                call_sign = [call_sign, ' '].join
            end

            return call_sign.chars.map { |p| p.ord << 1 } + [enc_ssid]
        end

        private
        def self.encode_frame(frame)
            enc_frame = encode_callsign(parse_identity_string(frame[:destination].to_s)) + encode_callsign(parse_identity_string(frame[:source].to_s))

            frame[:path].each do |hop|
                enc_frame += encode_callsign(parse_identity_string(hop.to_s))
            end

            return enc_frame[0...-1] + [enc_frame[-1] | 0x01] + [SLOT_TIME] + [0xf0] + frame[:payload].chars.map { |c| c.ord }
        end

        private
        def self.decode_frame(raw_frame)
            frame_len = raw_frame.length

            if frame_len > 16
                (0...frame_len - 2).each do |raw_slice|
                    # Is address field length correct?
                    if raw_frame[raw_slice] & 0x01 != 0 and ((raw_slice + 1) % 7) == 0
                        i = (raw_slice.to_f + 1.0) / 7.0
                        # Less than 2 callsigns?
                        if 1.0 < i and i < 11.0
                            if raw_frame[raw_slice + 1] & 0x03 == 0x03 and [0xf0, 0xcf].include? raw_frame[raw_slice + 2]
                                payload_as_array = raw_frame[raw_slice + 3..-1].map { |b| b.chr }
                                payload = payload_as_array.join
                                destination = identity_as_string(extract_callsign(raw_frame))
                                source = identity_as_string(extract_callsign(raw_frame[7..-1]))
                                path = extract_path(i.to_i, raw_frame)
                                return {:source => source, :destination => destination, :path => path, :payload => payload}
                            end
                        end
                    end
                end
            end
            return nil
        end

        private
        def self.valid_frame(raw_frame)
            frame_len = raw_frame.length

            if frame_len > 16
                (0...frame_len - 2).each do |raw_slice|
                    # Is address field length correct?
                    if raw_frame[raw_slice] & 0x01 != 0 and ((raw_slice + 1) % 7) == 0
                        i = (raw_slice.to_f + 1.0) / 7.0
                        # Less than 2 callsigns?
                        if 1.0 < i and i < 11.0
                            if raw_frame[raw_slice + 1] & 0x03 == 0x03 and [0xf0, 0xcf].include? raw_frame[raw_slice + 2]
                                return true
                            end
                        end
                    end
                end
            end
            return false
        end

        private
        def fill_buffer
            new_frames = []
            read_buffer = []
            read_data = read_interface
            while read_data and read_data.length > 0
                split_data = [[]]
                read_data.each do |read_byte|
                    if read_byte == FEND
                        split_data << []
                    else
                        split_data[-1] << read_byte
                    end
                end
                len_fend = split_data.length

                # No FEND in frame
                if len_fend == 1
                    read_buffer += split_data[0]
                    # Single FEND in frame
                elsif len_fend == 2
                    # Closing FEND found
                    if split_data[0]
                        # Partial frame continued, otherwise drop
                        new_frames << read_buffer + split_data[0]
                        read_buffer = []
                        # Opening FEND found
                    else
                        new_frames << read_buffer
                        read_buffer = split_data[1]
                    end
                    # At least one complete frame received
                elsif len_fend >= 3
                    (0...len_fend - 1).each do |i|
                        read_buffer_tmp = read_buffer + split_data[i]
                        if read_buffer_tmp.length > 0
                            new_frames << read_buffer_tmp
                            read_buffer = []
                        end
                    end
                    if split_data[len_fend - 1]
                        read_buffer = split_data[len_fend - 1]
                    end
                end
                # Get anymore data that is waiting
                read_data = read_interface
            end

            new_frames.each do |new_frame|
                if new_frame.length > 0 and new_frame[0] == 0
                    if @strip_df_start
                        new_frame = KissAbstract.strip_df_start(new_frame)
                    end
                    @frame_buffer << new_frame
                end
            end
        end

        private
        def read_bytes(*args, **kwargs)
              if @frame_buffer.length == 0
                  fill_buffer
              end

              if @frame_buffer.length > 0
                  return_frame = @frame_buffer[0]
                  @frame_buffer.shift
                  return return_frame
              else
                  return nil
              end
        end

        private
        def write_bytes(frame_bytes, port=0, *args, **kwargs)
              kiss_packet = [FEND] + [KissAbstract.command_byte_combine(port, DATA_FRAME)] +
                  KissAbstract.escape_special_codes(frame_bytes) + [FEND]

              write_interface(kiss_packet)
        end

        protected
        def write_setting(command, value)
            write_interface([FEND] + [command] + KissAbstract.escape_special_codes(value) + [FEND])
        end

        public
        def connect(mode_init=nil, *args, **kwargs)
        end

        public
        def close(*args, **kwargs)
        end

        public
        def read(*args, **kwargs)
          @lock.synchronize do
              frame = self.read_bytes(*args, **kwargs)
              if frame and frame.length > 0
                  return KissAbstract.decode_frame(frame)
              else
                  return nil
              end
          end
        end

        public
        def write(frame, port=0, *args, **kwargs)
          @lock.synchronize do
              encoded_frame = KissAbstract.encode_frame(frame)
              if KissAbstract.valid_frame(encoded_frame)
                self.write_bytes(encoded_frame, *args, **kwargs)
              else
                raise IOError.new("frame was able to be encoded but was determined not to be valid")
              end
          end
        end
    end
end
