require_relative '../../lib/kiss'
require 'abstractify'
require 'serialport'

ENCODED_FRAME = [192, 0, 158, 154, 142, 64, 64, 64, 96, 174, 100, 142, 154, 136, 64, 98, 174, 146, 136, 138, 98, 64, 98, 174, 146, 136, 138, 100, 64, 101, 3, 240, 116, 101, 115, 116, 95, 101, 110, 99, 111, 100, 101, 95, 102, 114, 97, 109, 101, 192]

DECODED_FRAME = {
                  :source => "W2GMD-1",
                  :destination => "OMG",
                  :payload => "test_encode_frame",
                  :path => ['WIDE1-1', 'WIDE2-2']
                }

describe Kiss::KissSerial do

  describe ".new" do
    context "Given valid arguments" do
      it "is successfully instantiated" do
        kiss_serial = Kiss::KissSerial.new("/dev/ttyUSB-scs")
        expect(kiss_serial).to_not be_nil
      end
    end
  end

  describe ".connect" do
    context "Given nil for mode_init, and no kwargs/extra arguments" do
      let( :serial_port ) { double(SerialPort) }
      let( :kiss_serial ) { Kiss::KissSerial.new("/dev/ttyUSB-scs") }
      it "sets read_timeout on underlying serial port" do
        allow( SerialPort ).to receive(:new).and_return(serial_port)
        expect(serial_port).to receive(:read_timeout=).with(-1).once
        kiss_serial.connect
      end
      it "does not call write_interface" do
        allow( SerialPort ).to receive(:new).and_return(serial_port)
        allow(serial_port).to receive(:read_timeout=).with(-1).once
        expect(kiss_serial).to_not receive(:write_interface)
        kiss_serial.connect
      end
      it "does not call write_setting" do
        allow( SerialPort ).to receive(:new).and_return(serial_port)
        allow(serial_port).to receive(:read_timeout=).with(-1).once
        expect(kiss_serial).to_not receive(:write_setting)
        kiss_serial.connect
      end
    end

    context "Given an arbitrary string for mode_init, and no kwargs/extra arguments" do
      let( :serial_port ) { double(SerialPort) }
      let( :kiss_serial ) { Kiss::KissSerial.new("/dev/ttyUSB-scs") }
      it "sets read_timeout on underlying serial port" do
        allow( SerialPort ).to receive(:new).and_return(serial_port)
        expect(serial_port).to receive(:read_timeout=).with(-1).once
        allow(serial_port).to receive(:write).with(Kiss::MODE_INIT_W8DED.map { |b| b.chr }.join)
        kiss_serial.connect(Kiss::MODE_INIT_W8DED)
      end
      it "calls write_interface" do
        allow( SerialPort ).to receive(:new).and_return(serial_port)
        allow(serial_port).to receive(:read_timeout=).with(-1).once
        expect(kiss_serial).to receive(:write_interface)
        allow(serial_port).to receive(:write).with(Kiss::MODE_INIT_W8DED.map { |b| b.chr }.join)
        kiss_serial.connect(Kiss::MODE_INIT_W8DED)
      end
      it "calls write_setting" do
        allow( SerialPort ).to receive(:new).and_return(serial_port)
        allow(serial_port).to receive(:read_timeout=).with(-1).once
        expect(kiss_serial).to_not receive(:write_setting)
        allow(serial_port).to receive(:write).with(Kiss::MODE_INIT_W8DED.map { |b| b.chr }.join)
        kiss_serial.connect(Kiss::MODE_INIT_W8DED)
      end
      it "calls write on underlying serial device with proper values" do
        allow( SerialPort ).to receive(:new).and_return(serial_port)
        allow(serial_port).to receive(:read_timeout=).with(-1).once
        expect(serial_port).to receive(:write).with(Kiss::MODE_INIT_W8DED.map { |b| b.chr }.join).once
        kiss_serial.connect(Kiss::MODE_INIT_W8DED)
      end
    end
  end

  describe ".close" do
    context "Given an underlying connection with mode_init set" do
      let( :serial_port ) { double(SerialPort) }
      let( :kiss_serial ) { Kiss::KissSerial.new("/dev/ttyUSB-scs") }

      it "ends KISS mode" do
        allow( SerialPort ).to receive(:new).and_return(serial_port)
        allow(serial_port).to receive(:read_timeout=)
        allow(serial_port).to receive(:write).with(Kiss::MODE_INIT_W8DED.map { |b| b.chr }.join)
        kiss_serial.connect(Kiss::MODE_INIT_W8DED)

        allow(serial_port).to receive(:closed?).and_return(false)
        allow(serial_port).to receive(:close)
        expect(serial_port).to receive(:write).with(Kiss::MODE_END.map { |b| b.chr }.join)
        kiss_serial.close
      end

      it "closes underlying serial port" do
        allow( SerialPort ).to receive(:new).and_return(serial_port)
        allow(serial_port).to receive(:read_timeout=)
        allow(serial_port).to receive(:write).with(Kiss::MODE_INIT_W8DED.map { |b| b.chr }.join)
        kiss_serial.connect(Kiss::MODE_INIT_W8DED)

        allow(serial_port).to receive(:closed?).and_return(false)
        expect(serial_port).to receive(:close)
        allow(serial_port).to receive(:write).with(Kiss::MODE_END.map { |b| b.chr }.join)
        kiss_serial.close
      end
    end

    context "Given an underlying connection without mode_init set" do
      let( :serial_port ) { double(SerialPort) }
      let( :kiss_serial ) { Kiss::KissSerial.new("/dev/ttyUSB-scs") }

      it "does not attempt to exit KISS mode" do
        allow( SerialPort ).to receive(:new).and_return(serial_port)
        allow(serial_port).to receive(:read_timeout=)
        kiss_serial.connect

        allow(serial_port).to receive(:closed?).and_return(false)
        allow(serial_port).to receive(:close)
        expect(serial_port).to_not receive(:write)
        kiss_serial.close
      end

      it "closes underlying serial port" do
        allow( SerialPort ).to receive(:new).and_return(serial_port)
        allow(serial_port).to receive(:read_timeout=)
        kiss_serial.connect

        allow(serial_port).to receive(:closed?).and_return(false)
        expect(serial_port).to receive(:close)
        kiss_serial.close
      end
    end

    context "Given an underlying connection that has been closed" do
      let( :serial_port ) { double(SerialPort) }
      let( :kiss_serial ) { Kiss::KissSerial.new("/dev/ttyUSB-scs") }

      it "closes underlying serial port" do
        allow( SerialPort ).to receive(:new).and_return(serial_port)
        allow(serial_port).to receive(:read_timeout=)
        kiss_serial.connect

        allow(serial_port).to receive(:closed?).and_return(true)
        expect(serial_port).to_not receive(:close)
        expect {
          kiss_serial.close
        }.to raise_error(IOError)
      end
    end
  end

  describe ".write" do
    context "Given an underlying connection and valid frame" do
      let( :serial_port ) { double(SerialPort) }
      let( :kiss_serial ) { Kiss::KissSerial.new("/dev/ttyUSB-scs") }

      it "encodes the frame properly" do
        allow( SerialPort ).to receive(:new).and_return(serial_port)
        allow(serial_port).to receive(:read_timeout=)
        allow(serial_port).to receive(:write).with(Kiss::MODE_INIT_W8DED.map { |b| b.chr }.join)
        kiss_serial.connect(Kiss::MODE_INIT_W8DED)

        expect(serial_port).to receive(:write).with(ENCODED_FRAME.map { |b| b.chr }.join).once
        kiss_serial.write(DECODED_FRAME)
      end
    end
  end

  describe ".read" do
    context "Given an underlying connection with a valid frame waiting" do
      let( :serial_port ) { double(SerialPort) }
      let( :kiss_serial ) { Kiss::KissSerial.new("/dev/ttyUSB-scs") }

      it "decodes the frame properly" do
        allow( SerialPort ).to receive(:new).and_return(serial_port)
        allow(serial_port).to receive(:read_timeout=)
        allow(serial_port).to receive(:write).with(Kiss::MODE_INIT_W8DED.map { |b| b.chr }.join)
        kiss_serial.connect(Kiss::MODE_INIT_W8DED)

        expect(serial_port).to receive(:read).and_return(ENCODED_FRAME.map { |b| b.chr }.join, nil)
        actual_frame = kiss_serial.read
        expect(actual_frame).to eql(DECODED_FRAME)
      end
    end
  end

end
